%% tclp__clp_types.pl
%% Copyright (C) 2001,2002 INRIA Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                  TCLP                                     %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/beaune/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__clp_types.pl,v 1.14 2003/09/12 11:17:10 coquery Exp $

:- module(tclp__clp_types, [':<'/2,
                            fresh/1,
                            fresh_term/1,
                            init/0,
                            types_reinit/0,
                            getbounds/3,
                            check_constraint/1,
                            term_type_up/2,
                            term_type_down/2,
                            cyclic/1, cyclic_up/1, cyclic_down/1,
                            type_info/2,
                            get_greater_vars/2,
                            get_lesser_vars/2,
                            try_generalize_var/2,
                            ordered_vars/2,
			    is_type_variable/1,
                            tclp__enumerate_bounds/0,
                            try_bound_enumeration/0
                            ]).

:- use_module(tclp(tclp__def)).
:- use_module(tclp(tclp__errors)).
:- use_module(tclp(tclp__type_order)).
:- use_module(library(chr)).
:- use_module(library(ordsets)).
:- use_module(library(lists)).
:- use_module(tclp(tclp__variables_set_assoc)).
:- use_module(tclp(tclp__utils)).

%% Verbose modes: debug_clp_types

handler type.

constraints
            %% subtyping constraint
            (':<')/2  , ('delayed:<')/2,
            tclp__var_ineq/2 , tclp__hi/2, tclp__lo/2,
            %% expressing type constraints over a type variable
            tclp__parameter/5,
            %% updates the hiset and hibounds (resp. loset and lobound)
            %% of a list of variables
            tclp__update_hiset/2, tclp__update_loset/2,
            %% updates the lower (resp. greater) bound of some variables
            tclp__update_lobound/2, tclp__update_hibound/2,
            %% stops constraints until updates are done
            tclp__sem/0,
            %% store the set of original leaves of a variable
            tclp__original_up/2, tclp__original_down/2,
            %% computes glbs and lubs of variables
            tclp__vGLB/3, tclp__vLUB/3,
            %% make the break
            tclp__do_break/1,
            %% marker for original variables
            tclp__original/1,
            %% marker for the presence of a variable in some bound
            tclp__in_up/2,
            tclp__in_down/2,
            tclp__add_in_up/2,
            tclp__add_in_down/2,
            tclp__rem_in_up/2,
            tclp__rem_in_down/2,
            %% triggers the fact that the constrained variable is equal
            %% to top (resp to bottom)
            tclp__trigger_top/1,
            tclp__trigger_bottom/1,
            %% elim doqn and elim up
            tclp__elim_down/2,
            tclp__elim_up/2,
            %% bound enumeration
            tclp__enumerate_bounds/0
            .

operator(100,xfx,[':<']).

%%%%%%%%%%%%%%%%%%%%%%%%%
%% Solver's core rules %%
%%%%%%%%%%%%%%%%%%%%%%%%%

%% error if a type variable is instanciated
tclp__parameter(X,_,_,_,_) <=>
  nonvar(X) | tclp__variables_set_assoc:print,
              chr:findall_constraints(tclp__original_up(X,_),L),
              write(L),nl,nl,
              print_error([ 'instanciated variable ',X]).

%% trigger bottom / top elimination
tclp__parameter(X,Hibound,_Hiset,_Loset,_Lobound) ==>
  builtin_type(bottom,Bot), Hibound == Bot |
  tclp__trigger_bottom(X).
tclp__parameter(X,_Hibound,_Hiset,_Loset,Lobound) ==>
  builtin_type(top,Top), Lobound == Top |
  tclp__trigger_top(X).

%% failure when triggering original variables
tclp__trigger_top(X), tclp__original(X) <=> fail.
tclp__trigger_bottom(X), tclp__original(X) <=> fail.

%% trigger elimination on all concerned variables
tclp__trigger_elimination_down(_X,[]).
tclp__trigger_elimination_down(X,[Y|Rest]) :-
        tclp__elim_down(X,Y),
        tclp__trigger_elimination_down(X,Rest).
tclp__trigger_elimination_up(_X,[]).
tclp__trigger_elimination_up(X,[Y|Rest]) :-
        tclp__elim_up(X,Y),
        tclp__trigger_elimination_up(X,Rest).

%% rules to deal with introduced variables forced to top/or bottom
tclp__trigger_bottom(X), tclp__in_up(X,List)
        <=> tclp__in_up(X,[]),
            tclp__trigger_elimination_down(X,List).
tclp__trigger_top(X), tclp__in_down(X,List)
        <=> tclp__in_down(X,[]),
            tclp__trigger_elimination_up(X,List).


%% down rule
tclp__parameter(Y,Hibound,_Hiset,_Loset,_Lobound) \
tclp__elim_down(X,Y) 
        <=> Hibound =.. [_| Args],
            true_nth(N,Args,X),
            tclp__elim_arg_down(Hibound,N,NewHibound),
            Y :< NewHibound.

%% up rule 
tclp__parameter(Y,_Hibound,_Hiset,_Loset,Lobound) \
tclp__elim_up(X,Y)
        <=> Lobound =.. [_| Args],
            true_nth(N,Args,X),
            tclp__elim_arg_up(Lobound,N,NewLobound),
            NewLobound :< Y.

% when a break is needed, do it.

tclp__parameter(X, Hibound, Hiset, Loset, Lobound) \
tclp__do_break(X)
       <=> tclp__break(Lobound, Hibound).

% combinaison of two sets of constraints over the same variable
% generaly used after unification of two type variables
tclp__parameter(X, Hibound1, Hiset1, Loset1, Lobound1),
tclp__parameter(X, Hibound2, Hiset2, Loset2, Lobound2),
tclp__sem
       <=> verbose(debug_clp_types,'clp types: unifying ...'),
           list_to_ord_set(Hiset1,SHiset1),
           list_to_ord_set(Hiset2,SHiset2),
           list_to_ord_set(Loset1,SLoset1),
           list_to_ord_set(Loset2,SLoset2),
           ord_union(SHiset1, SHiset2, NH),
           ord_del_element(NH, X,NewHiset),
           ord_union(SLoset1, SLoset2, NL),
           ord_del_element(NL,X,NewLoset),
           tclp__update_hiset_s(NewLoset,NewHiset),
           tclp__update_loset_s(NewHiset,NewLoset),
           Hibound1 =.. [_|OldArgsh1], unregister_up(X,OldArgsh1),
           Hibound2 =.. [_|OldArgsh2], unregister_up(X,OldArgsh2),
           tclp__glb(Hibound1,Hibound2,NewHibound),
           NewHibound =.. [_|NewArgsh], register_up(X,NewArgsh),
           Lobound1 =.. [_|OldArgsl1], unregister_down(X,OldArgsl1),
           Lobound2 =.. [_|OldArgsl2], unregister_down(X,OldArgsl2),
           tclp__lub(Lobound1,Lobound2,NewLobound),
           NewLobound =.. [_|NewArgsl], register_down(X,NewArgsl),
           tclp__parameter(X,NewHibound,NewHiset,NewLoset,NewLobound),
           tclp__update_hibound_s(NewLoset,NewHibound),
           tclp__update_lobound_s(NewHiset,NewLobound),
           tclp__do_break(X),
           verbose(debug_clp_types,'clp types: unifying ok'),
           tclp__sem.

%% deal with unification for tclp__in_up/tclp__in_down
tclp__in_up(X,L1), tclp__in_up(X,L2)
       <=> get_ordset(L1,O1), get_ordset(L2,O2),
           ord_union(O1,O2,O3), tclp__in_up(X,O3).
tclp__in_down(X,L1), tclp__in_down(X,L2)
       <=> get_ordset(L1,O1), get_ordset(L2,O2),
           ord_union(O1,O2,O3), tclp__in_down(X,O3).


%% delay the posting of a constraint if we don't the nature of one of
%% the arguments
X :< Y
       <=> (var(X), \+ is_type_variable(X) ; var(Y), \+ is_type_variable(Y))
       | 'delayed:<'(X,Y).

%% reconsider a delayed constraint
tclp__parameter(X,_,_,_,_),tclp__parameter(Y,_,_,_,_)
  \ 'delayed:<'(X,Y)
       <=> X:<Y.
tclp__parameter(X,_,_,_,_) \ 'delayed:<'(X,Y) <=> nonvar(Y)|X:<Y.
tclp__parameter(Y,_,_,_,_) \ 'delayed:<'(X,Y) <=> nonvar(X)|X:<Y.
'delayed:<'(X,Y) <=> nonvar(X), nonvar(Y) | X:<Y.

% posing constraints :<
% transform the generic constraint :< to more specific ones :
% between variables, or constraint a variable by a lower (resp. greater) bound
X :< Y , tclp__sem <=> var(X),var(Y) |
                 tclp__var_ineq(X,Y), tclp__sem.
X :< Y , tclp__sem <=> nonvar(X),var(Y) |
                 tclp__transinf(X,XFlat) , tclp__lo(XFlat,Y), tclp__sem.
X :< Y , tclp__sem <=> var(X), nonvar(Y) |
                 tclp__transup(Y,YFlat) , tclp__hi(X,YFlat), tclp__sem.
X :< Y , tclp__sem <=> nonvar(X), nonvar(Y) |
                 tclp__transinf(X,XFlat) , tclp__transup(Y,YFlat),
                 tclp__break(XFlat, YFlat), tclp__sem.

% X :< X 
tclp__var_ineq(X,X) <=> true.

% X :< Y , changes the constraints over X and Y
% (expressed by the tclp__parameter constraint)
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX),
tclp__parameter(Y, HiboundY, HisetY, LosetY, LoboundY) \
tclp__var_ineq(X,Y)
       <=> true |
           verbose(debug_clp_types,['clp types: ',X,' < ',Y,' ']),
           list_to_ord_set(HisetX,SHisetX),
           (
             ord_member(Y,SHisetX) ->
             verbose(debug_clp_types,
                     ['clp types: ',X,' < ',Y,' already done']),true
                                % the constraint was already in the store
           ;
             (
               list_to_ord_set(LosetX, SLosetX),
               ord_add_element(SLosetX,X,Loset),
                                % the variables to add to Y and its hiset
               list_to_ord_set(HisetY,SHisetY),
               ord_add_element(SHisetY,Y,Hiset),
                                % the variables to add to X and its loset
               tclp__update_hiset_s(Loset, Hiset),
               tclp__update_loset_s(Hiset, Loset),
               tclp__update_hibound_s(Loset, HiboundY),
               tclp__update_lobound_s(Hiset, LoboundX),
                                % check compatibility of the bounds of X and Y
                                % one could check bound compatibility on one
                                % of X or Y since LoboundX < LoboundY
               tclp__break(LoboundX, HiboundY)
             ),
             verbose(debug_clp_types,['clp types: ',X,' < ',Y,' ok'])
           ).
           

% X :< tau
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX) \
tclp__hi(X,Hibound)
       <=> verbose(debug_clp_types,[X,':<',Hibound]),
           tclp__update_hibound_s([ X | LosetX ], Hibound),
           tclp__do_break(X).

% tau :< X
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX) \
tclp__lo(Lobound,X)
       <=> tclp__update_lobound_s([ X | HisetX ], Lobound),
           tclp__do_break(X).

%%%%%%%%%
%% updating parameters

%% change X's hibound and hiset

tclp__update_hiset(X,[]) <=> true. % no point in updating with an empty hiset

tclp__update_hiset(X, Hiset) ,
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX)
       <=> % Hiset is already a ordered set
           list_to_ord_set(HisetX, SHisetX),
           ord_union(Hiset, SHisetX, NewHiset),
           tclp__parameter(X, HiboundX, NewHiset, LosetX, LoboundX).

% do not change X's hibound if it is already smaller than the new bound
tclp__parameter(X, HiboundX, _, _, _) \
tclp__update_hibound(X,Hibound)
       <=> tclp__ordered_types(HiboundX, Hibound) | true.

% change X's hibound 
tclp__update_hibound(X,Hibound) ,
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX)
       <=> tclp__glb(HiboundX, Hibound, NewHibound),
           HiboundX =.. [_|OldArgs], unregister_up(X,OldArgs),
           NewHibound =.. [_|NewArgs], register_up(X,NewArgs),
           tclp__parameter(X, NewHibound, HisetX, LosetX, LoboundX).

% change X's loset
tclp__update_loset(X,[]) <=> true. % no point in updating with an empty loset
tclp__update_loset(X, Loset) ,
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX)
       <=> % Loset is already a ordered set
           list_to_ord_set(LosetX, SLosetX),
           ord_union(Loset, SLosetX, NewLoset),
           tclp__parameter(X, HiboundX, HisetX, NewLoset, LoboundX).


% do not change X's lobound if it is already greater than the new bound
tclp__parameter(X, _, _, _, LoboundX) \
tclp__update_lobound(X,Lobound)
       <=> tclp__ordered_types(Lobound, LoboundX) | true.

% change X's lobound 
tclp__update_lobound(X,Lobound) ,
tclp__parameter(X, HiboundX, HisetX, LosetX, LoboundX)
       <=> tclp__lub(LoboundX, Lobound, NewLobound),
           LoboundX =.. [_|OldArgs], unregister_down(X,OldArgs),
           NewLobound =.. [_|NewArgs], register_down(X,NewArgs),
           tclp__parameter(X, HiboundX, HisetX, LosetX, NewLobound).

%%%%%%%%%
%% compute a lub

%% LUB(X,X) is X
tclp__vLUB(X,X,LUB) <=>  LUB = X.

%% both X and Y are introduced variables
tclp__original_down(X,OrigX),
tclp__original_down(Y,OrigY) \
tclp__vLUB(X,Y,LUB)
       <=> list_to_ord_set(OrigX,SOrigX),
           list_to_ord_set(OrigY,SOrigY),
           ord_union(SOrigX,SOrigY,OrigLUB),
           tclp__LUBVar(OrigLUB,LUB),
           X:<LUB, Y:<LUB.

%% the second variable is an original
tclp__original_down(X,OrigX) \
tclp__vLUB(X,Y,LUB)
       <=> list_to_ord_set(OrigX,SOrigX),
           ord_add_element(SOrigX,Y,OrigLUB),
           tclp__LUBVar(OrigLUB,LUB),
           X:<LUB, Y:<LUB.

%% the first variable is an original : we switch the two variables
%% to apply the preceeding rule
tclp__original_down(X,OrigX) \
tclp__vLUB(Y,X,LUB) <=> tclp__vLUB(X,Y,LUB).

%% both variables are originals
tclp__vLUB(X,Y,LUB) <=>
       list_to_ord_set([X,Y],OrigLUB),
       tclp__LUBVar(OrigLUB,LUB),
       X:<LUB, Y:<LUB.

%%%%%%%%%
%% compute a glb

%% GLB(X,X) is X
tclp__vGLB(X,X,GLB) <=> GLB = X.

%% both X and Y are introduced variables
tclp__original_up(X,OrigX),
tclp__original_up(Y,OrigY) \
tclp__vGLB(X,Y,GLB)
       <=> list_to_ord_set(OrigX,SOrigX),
           list_to_ord_set(OrigY,SOrigY),
           ord_union(SOrigX,SOrigY,OrigGLB),
           tclp__GLBVar(OrigGLB,GLB),
           GLB:<X, GLB:<Y.

%% the second variable is an original
tclp__original_up(X,OrigX) \
tclp__vGLB(X,Y,GLB)
       <=> list_to_ord_set(OrigX,SOrigX),
           ord_add_element(SOrigX,Y,OrigGLB),
           tclp__GLBVar(OrigGLB,GLB),
           GLB:<X, GLB:<Y.

%% the first variable is an original: we switch the two variables
% to apply the preceeding rule
tclp__original_up(X,OrigX) \
tclp__vGLB(Y,X,GLB) <=> tclp__vGLB(X,Y,GLB).

%% both variables are originals
tclp__vGLB(X,Y,GLB) <=>
       list_to_ord_set([X,Y],OrigGLB),
       tclp__GLBVar(OrigGLB,GLB),
       GLB:<X, GLB:<Y.


%%%%%%%%%
%% dealing with original variables
tclp__original(X) \ tclp__original(X) <=> true.

%%%%%%%%%
%% dealing with variables that have no solution 

%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Auxiliary predicates %%
%%%%%%%%%%%%%%%%%%%%%%%%%%

% calling constraints over a list of a variable

tclp__update_hiset_s([],_).
tclp__update_hiset_s([X|L],Hiset) :-
        tclp__update_hiset(X, Hiset),
        tclp__update_hiset_s(L, Hiset).

tclp__update_loset_s([],_).
tclp__update_loset_s([X|L], Loset) :-
        tclp__update_loset(X, Loset),
        tclp__update_loset_s(L, Loset).

tclp__update_hibound_s([],_).
tclp__update_hibound_s([X|L],Hibound) :-
        tclp__update_hibound(X,Hibound),
        tclp__update_hibound_s(L,Hibound).

tclp__update_lobound_s([],_).
tclp__update_lobound_s([X|L],Lobound) :-
        tclp__update_lobound(X,Lobound),
        tclp__update_lobound_s(L,Lobound).

% transforming a type term into a flat one with variables
% tclp__transup does it by constraining up variables

% a variable is always flat
tclp__transup(X,X) :- var(X),!.

% compound type
tclp__transup(Type,Flat) :-
        Type =.. [Constr|Args],
        tclp__transup_args(Args,FlatArgs),
        Flat =.. [Constr|FlatArgs].

% for args list
tclp__transup_args([],[]).
tclp__transup_args([T|L],[X|R]) :-
        (var(T) -> X=T ; (fresh(X) ,X :< T)),
        tclp__transup_args(L,R).

% tclp__transinf does it by constraining down variables

% a variable is always flat
tclp__transinf(X,X) :- var(X),!.

% compound type
tclp__transinf(Type,Flat) :-
        Type =.. [Constr|Args],
        tclp__transinf_args(Args,FlatArgs),
        Flat =.. [Constr|FlatArgs].

% for args list
tclp__transinf_args([],[]).
tclp__transinf_args([T|L],[X|R]) :-
        (var(T) -> X=T ; (fresh(X) ,T :< X)),
        tclp__transinf_args(L,R).

%% tests if a variable is a type variable
is_type_variable(X) :-
	var(X),!,chr:find_constraint(X,tclp__parameter(X,_,_,_,_),_).

% creating new type variables
core_fresh(X) :- var(X), !,
        builtin_type(top,Top),
        builtin_type(bot,Bot),
        tclp__parameter(X,Top,[],[],Bot),
        tclp__in_down(X,[]),
        tclp__in_up(X,[]).
%% if a variable is already a type variable, no point doing a new fresh
fresh(X) :- is_type_variable(X),!.
%% standard fresh ...
fresh(X) :-
        core_fresh(X),
        register(X),
        tclp__original(X).

%% fresh_term(Term)
%% call fresh on all the variables of term
fresh_term(X) :- var(X), !, fresh(X).
fresh_term([]) :- !.
fresh_term([X|Rest]) :- !, fresh_term(X), fresh_term(Rest).
fresh_term(Term) :- Term =.. [ _ | Args ],!, fresh_term(Args).

tclp__GLBVar(OrigGLB,GLB) :-
	option(usetreelookup) ->
	tclp__GLBVar2(OrigGLB,GLB);
	tclp__GLBVar1(OrigGLB,GLB).

tclp__GLBVar1(OrigGLB,GLB) :-
        %OrigGLB is sorted
        chr:find_constraint(tclp__original_up(GLB,ConstraintOrigs),_),
        list_to_ord_set(ConstraintOrigs,ConstraintOrigsToCompare),
        ord_seteq(ConstraintOrigsToCompare, OrigGLB),!.

tclp__GLBVar1(OrigGLB,GLB) :-
        %OrigGLB is sorted
        tclp__original_down(GLB,[]),
        tclp__original_up(GLB,OrigGLB),
        core_fresh(GLB),
        verbose(debug_clp_types,['clp types: created glb ',GLB,' for ',OrigGLB]),
        tclp__link_up(GLB,OrigGLB).

tclp__GLBVar2(OrigGLB,GLB) :-
        %OrigGLB is sorted
        associated(OrigGLB,glblub(GLB,_)),
        (
          chr:find_constraint(GLB,tclp__original_up(GLB,_),_)
        ->
          true
        ;
          tclp__original_down(GLB,[]),
          tclp__original_up(GLB,OrigGLB),
          core_fresh(GLB),
          verbose(debug_clp_types,
                  ['clp types: created glb ',GLB,' for ',OrigGLB]),
          tclp__link_up(GLB,OrigGLB)
        ).
        
tclp__link_up(_,[]).
tclp__link_up(X,[Y|Ys]) :-
        X :< Y,
        tclp__link_up(X,Ys).

tclp__LUBVar(OrigLUB,LUB) :-
	option(usetreelookup) ->
	tclp__LUBVar2(OrigLUB,LUB);
	tclp__LUBVar1(OrigLUB,LUB).

tclp__LUBVar1(OrigLUB,LUB) :-
        % OrigLUB is sorted
        chr:find_constraint(tclp__original_down(LUB,ConstraintOrigs),_),
        list_to_ord_set(ConstraintOrigs,ConstraintOrigsToCompare),
        ord_seteq(ConstraintOrigsToCompare, OrigLUB),!.
tclp__LUBVar1(OrigLUB,LUB) :-
        tclp__original_up(LUB,[]),
        tclp__original_down(LUB,OrigLUB),
        core_fresh(LUB),
        verbose(debug_clp_types,['clp types: created lub ',LUB,' for ',OrigLUB]),
        tclp__link_down(OrigLUB,LUB).

tclp__LUBVar2(OrigLUB,LUB) :-
        %OrigLUB is sorted
        associated(OrigLUB,glblub(_,LUB)),
        (
          chr:find_constraint(LUB,tclp__original_down(LUB,_),_)
        ->
          true
        ;
          tclp__original_up(LUB,[]),
          tclp__original_down(LUB,OrigLUB),
          core_fresh(LUB),
          verbose(debug_clp_types,
                  ['clp types: created lub ',LUB,' for ',OrigLUB]),
          tclp__link_down(OrigLUB,LUB)
        ).


tclp__link_down([],_).
tclp__link_down([X|Xs],Y) :-
        X :< Y,
        tclp__link_down(Xs,Y).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Dealing with the registration of introduced variables %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% register X as a variable to wake 
register_up(_X,[]).
register_up(X,[Y|Rest]) :-
        tclp__add_in_up(X,Y),
        register_up(X,Rest).
register_down(_X,[]).
register_down(X,[Y|Rest]) :-
        tclp__add_in_down(X,Y),
        register_up(X,Rest).

%% unregister
unregister_down(_X,[]).
unregister_down(X,[Y|Rest]) :-
        tclp__rem_in_down(X,Y),
        unregister_down(X,Rest).
unregister_up(_X,[]).
unregister_up(X,[Y|Rest]) :-
        tclp__rem_in_up(X,Y),
        unregister_up(X,Rest).

%% rules for registration update
tclp__in_up(Y,List) \ tclp__add_in_up(X,Y) <=> true_member(X,List) | true.
tclp__in_up(Y,List), tclp__add_in_up(X,Y) <=> tclp__in_up(Y,[X|List]).

tclp__in_down(Y,List) \ tclp__add_in_down(X,Y) <=> true_member(X,List) | true.
tclp__in_down(Y,List), tclp__add_in_down(X,Y) <=> tclp__in_down(Y,[X|List]).

tclp__in_up(Y,List), tclp__rem_in_up(X,Y)
        <=> delete(List,X,NewList),tclp__in_up(Y,NewList).
tclp__in_down(Y,List), tclp__rem_in_down(X,Y)
        <=> delete(List,X,NewList),tclp__in_down(Y,NewList).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% enumeration of bounds
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% enumerate possible tops for a type variable which is not upper bounded
tclp__enumerate_bounds,
tclp__parameter(X,Hibound,Hiset,Loset,Lobound) \
tclp__sem <=> builtin_type(top,Top), Hibound==Top,
                                % the hibound is not fixed
              builtin_type(bottom,Bot), Lobound\=Bot
                                % here put appart completely unbounded variable
              |
              (tclp__enumerate_tops(Lobound) ->
                                % if Lobound is a maximal element,
                                % we use it as the hibound
                                % since it is the only one possible
                                % This optimize well the case of constants used
                                % for testing definitionnal genericity
                  incr_counter(bound_tries,_,_), X :< Lobound
              ;
                  tclp__enumerate_tops(SomeTop),
                  incr_counter(bound_tries,_,_), X :< SomeTop ),
              tclp__sem.

%% enumerate possible tops for a type variable which is not lower bounded
tclp__enumerate_bounds,
tclp__parameter(X,Hibound,_Hiset,_Loset,Lobound) \
tclp__sem <=> builtin_type(bottom,Bot), Lobound==Bot,
                                % the lobound is not fixed
              builtin_type(top,Top), Hibound\=Top
                                % here put appart completely unbounded variable
              |
              (tclp__enumerate_bottoms(Hibound) ->
                                % if Hibound is a minimal element,
                                % we use it as the lobound
                                % since it is the only one possible
                                % This optimize well the case of constants used
                                % for testing definitionnal genericity
                  incr_counter(bound_tries,_,_), Hibound :< X ;
                  tclp__enumerate_bottoms(SomeBot), incr_counter(bound_tries,_,_), SomeBot :< X ),
              tclp__sem.

%% if there is no more enumeration to do, the constraint tclp__enumerate_bounds
%% disappears
tclp__sem \ tclp__enumerate_bounds <=> true.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% (Re)Initialisation of the constraint store
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

init :- chr:findall_constraints(tclp__sem, []),!,
        tclp__sem.
init.

%% reinit
%% removes all the constraints in the type checker solver
remove_cstr(_ # Id) :- chr:remove_constraint(Id).
remove_cstrs([]).
remove_cstrs([X|L]) :-
        remove_cstr(X), remove_cstrs(L).
rem_constraints(Pattern) :-
        chr:findall_constraints(Pattern, Constraints),
        remove_cstrs(Constraints).
types_reinit.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Test of satisfiability of some constraints without effective commit
%% (using backtracking)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% check(constraint)
%% Testing of a new constraint without commiting it to the current
%% constraint store
check_constraint(C) :- %\+ \+ C.
        findall(R,(C,!,R=a),L),!,
        L=[a].

%% try to enumerate possible bounds for all type variables in order to test
%% the constraints satisfiability
try_bound_enumeration :-
        init_counter(bound_tries),
        check_constraint(tclp__enumerate_bounds),
        (option(print_stats) ->
            current_counter_value(bound_tries,N),
            write(N),write(' bounds tried'),nl;
            true).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Information on type variables
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%
%% Bounds infos %%
%%%%%%%%%%%%%%%%%%

%% Get the upper/lower bounds of a type variable
%% :- pred getbounds(type_variable, type_term, type_term).
getbounds(X,UB,LB) :-
        var(X),!,
        once((
              chr:findall_constraints(X, tclp__parameter(X,_,_,_,_),
                                      Constraints),
              Constraints = [tclp__parameter(X,UB,_,_,LB)#_|_]
             ;
              print_error(['Can''t get type constraints for ',X]),fail
             )).
getbounds(X,_,_) :-
        print_error(['Type variable was instanciated : ',X]),fail.

%% Get some infos about a type variable
%% :- pred type_info(type_variable, pair(type_term, type_term)).
type_info(X,LB-UB) :-
        cyclic(X),!,LB = UB, UB = cycle.
type_info(X,LB-UB) :-
        getbounds(X,UB1,LB1),!,
        UB1 =.. [ UTCons | UArgs ],
        u_type_infos(UArgs, UArgsInfos),
        UB =.. [ UTCons | UArgsInfos ],
        LB1 =.. [ LTCons | LArgs ],
        b_type_infos(LArgs, LArgsInfos),
        LB =.. [ LTCons | LArgsInfos ]
        ;
        print_error('type info failed'),fail.
type_info(_,_) :-
        print_error('get_bounds failed'),fail.

u_type_infos([],[]).
u_type_infos([X|Xs],[I|Is]) :-
        type_info(X,_-I),
        u_type_infos(Xs,Is).

b_type_infos([],[]).
b_type_infos([X|Xs],[I|Is]) :-
        type_info(X,I-_),
        b_type_infos(Xs,Is).

all_term_up([],[]).
all_term_up([X|Rest],[UB|BRest]) :-
        term_type_up(X,UB),
        all_term_up(Rest,BRest).

all_term_down([],[]).
all_term_down([X|Rest],[LB|BRest]) :-
        term_type_down(X,LB),
        all_term_down(Rest,BRest).

term_type_up(X,UB) :-
        (
          var(X)
        ->
	  fresh(X),
          type_info(X,_-UB)
        ;
          X =.. [Head|Args],
          all_term_up(Args,ArgsB),
          UB =.. [Head|ArgsB]%,
        ).

term_type_down(X,LB) :-
        (
          var(X)
        ->
	  fresh(X),
          type_info(X,LB-_)
        ;
          X =.. [Head|Args],
          all_term_down(Args,ArgsB),
          LB =.. [Head|ArgsB]
        ).

%%%%%%%%%%%%%%%%%%
%% Cycles infos %%
%%%%%%%%%%%%%%%%%%

%% Tests if there is a cycle in some term
cyclic_up(A) :-
        cyclic_up(A,[],[],Ups),
        ord_member(A,Ups).
cyclic_up(A,Done,Current,Result) :-
        (
          ord_member(A,Done)
        ->
          ord_add_element(Current,A,Result)
        ;
          getbounds(A,_,UB),
          UB =.. [ _ | UArgs ],
          ord_add_element(Done,A,Done2),
          cyclic_up2(UArgs,Done2,Current,Result)
        ).
cyclic_up2([],_,Result,Result).
cyclic_up2([A|Rest],Done,Current,Result) :-
        cyclic_up(A,Done,Current,Accu),
        cyclic_up2(Rest,Done,Accu,Result).

cyclic_down(A) :-
        cyclic_down(A,[],[],Downs),
        ord_member(A,Downs).
cyclic_down(A,Done,Current,Result) :-
        (
          ord_member(A,Done)
        ->
          ord_add_element(Current,A,Result)
        ;
          getbounds(A,LB,_),
          LB =.. [ _ | UArgs ],
          ord_add_element(Done,A,Done2),
          cyclic_down2(UArgs,Done2,Current,Result)
        ).
cyclic_down2([],_,Result,Result).
cyclic_down2([A|Rest],Done,Current,Result) :-
        cyclic_down(A,Done,Current,Accu),
        cyclic_down2(Rest,Done,Accu,Result).

cyclic(A) :- cyclic_up(A) ; cyclic_down(A).

%%%%%%%%%%%%%%%%%
%% Order infos %%
%%%%%%%%%%%%%%%%%

%% Vars are all the type variables Y such that X :< Y
get_greater_vars(X, Vars) :-
        chr:find_constraint(X,tclp__parameter(X, _, Vars, _, _),_),!.

%% Vars are all the type variables Y such that Y :< X
get_lesser_vars(X, Vars) :-
        chr:find_constraint(X,tclp__parameter(X, _, _, Vars, _),_),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Propagation of the generalisation of a variable
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

strict_member(_,[]) :- fail.
strict_member(X,[Y|_]) :- X == Y,!.
strict_member(X,[_|L]) :- strict_member(X,L).

%% Finds all variables that are linked to X
find_linked(X,Current,All) :-
        (
          strict_member(X,Current)
        ->
          Current = All
        ;
          get_greater_vars(X, Greater),!,
          find_all_linked(Greater, [ X | Current ], Inter),
          get_lesser_vars(X, Lesser),!,
          find_all_linked(Lesser, Inter, All),!
        ).
find_all_linked([],All,All).
find_all_linked([X|L],Current,All) :-
        find_linked(X,Current,Inter),
        find_all_linked(L,Inter,All).

%%% Try the generalisation of X to the constant Name
%try_generalize_var(X, Name, Current, Done) :-
%        (
%          ord_member(X,Current)
%        ->
%          Current = Done
%        ;
%%          verbose(debug_type_inference,[ 'trying generalize ',X,' to ',Name]),
%          X :< Name, Name :< X,!,
%          ord_add_element(Current,X,Inter1),
%          get_greater_vars(X, Greater),!,
%%          verbose(debug_type_inference,[ 'uppers are ',Greater ]),
%          try_generalize_all_vars(Greater, Name, Inter1, Inter2),
%          get_lesser_vars(X, Lesser),!,
%%          verbose(debug_type_inference,[ 'lesser are ',Lesser ]),
%          try_generalize_all_vars(Lesser, Name, Inter2, Done),!
%        ).

%try_generalize_all_vars([], _, Done,Done).
%try_generalize_all_vars([ X | Xs ], Name, Current, Done) :-
%        try_generalize_var(X,Name,Current,Inter),!,
%        try_generalize_all_vars(Xs,Name,Inter,Done).

try_generalize_var2(X,Name) :-
        X :< Name, Name :< X,!.

try_generalize_all_vars([],_).
try_generalize_all_vars([X|Rest],Name) :-
        once(try_generalize_var2(X,Name)),
        try_generalize_all_vars(Rest,Name).

try_generalize_var(X,Name) :-
        find_linked(X,[],Linked),
        length(Linked,Len),
        verbose(debug_type_inference,['found ',Len,' vars linked to ',X]),
        try_generalize_all_vars(Linked,Name).

%% ordered_vars(X,Y) is true if X:<Y is already in the store
ordered_vars(X,Y) :-
        chr:find_constraint(X,tclp__parameter(X, _,Hiset,_,_),_),
        ord_member(Y,Hiset).

